#ifndef lint
static char vcid[] = "$Id$";
#endif

#include "sut.h"

/*
   The various pexec parts need to exchange a variety of messages.
   In order to provide the most flexible implementation, we use
   an active-message approach (though we essentially poll for the
   messages).  Uses can define their own messages, making it
   easy to customize the use of these routines for a variety of
   distributed computing tasks.

   All user-functions have the form

   int fcn( SUT_Jobs *, int index, SUT_Header *msg, void *user_ctx )

   index gives the connection; children, local process, and parent
   should all have jobs->conn entries.
 */
/****************************************************************************
 Here are basic routines to handle the core commands.  In some cases, 
 there will be different routines for the root and the intermediate
 roots (because the intermediate roots talk to each other, but the
 final root talks to the user). 
 ****************************************************************************/

/* Read data from a child process and send it to the parent */
int SUT_Data_from_child( jobs, i, msg, ctx )
SUT_Jobs   *jobs;
int        i;
SUT_Header *msg;
void       *ctx;
{
int err;
SUT_Connection *conn = jobs->conn + i;

if (SUT_Debug & SUT_PASSDATA)
    fprintf( stderr, "[%d] Reading %d bytes from %d\n", 
	    SUT_Mynode, msg->len, conn->nodenum );
if (msg->len > conn->linelen - conn->linepos) {
    fprintf( stderr, "Too much data; internal error\n" );
    msg->len = conn->linelen - conn->linepos;
    }
if (err = SUT_Readdata( conn->readfd, conn->linebuf + conn->linepos, 
		        msg->len )) {
    SUT_Errormessage( stderr, err );
    }
conn->linepos += msg->len;
(*jobs->writefilter)( jobs, i, 1 );
/* the filter routine is responsible for resetting linepos */
return SUT_SUCCESS;
}

#ifdef FOO
/* Read data from a parent process and send it to the children.
   Eventually, this will need to implement flow-control to the 
   children.

   i is the index of the PARENT in the jobs structure
 */
int SUT_Data_from_parent( jobs, i, msg, ctx )
SUT_Jobs   *jobs;
int        i;
SUT_Header *msg;
void       *ctx;
{
int  err, len;
char buf[SUT_MAX_BUF];
SUT_Connection *conn = jobs->conn + i;
int  j;

if (SUT_Debug & SUT_PASSDATA)
    fprintf( stderr, "[%d] Reading %d bytes from parent\n", 
	    SUT_Mynode, msg->len );
if (msg->len > conn->linelen - conn->linepos) {
    fprintf( stderr, "Too much data; internal error\n" );
    msg->len = conn->linelen - conn->linepos;
    }
if (err = SUT_Readdata( conn->readfd, conn->linebuf + conn->linepos, 
		        msg->len )) {
    SUT_Errormessage( stderr, err );
    }
conn->linepos += msg->len;
for (j=0; j<jobs->nchildren; j++) {
    conn = jobs->conn + jobs->children[j];
    if (conn->writefd >= 0) {
	}
    }
(*jobs->writefilter)( jobs, i, 1 );
/* the filter routine is responsible for resetting linepos */
return SUT_SUCCESS;
}
#endif

/* Handle a process(es)-exited message from a child; returns a 
   you-may-exit message */
int SUT_Do_Done( jobs, i, msg, ctx )
SUT_Jobs   *jobs;
int        i;
SUT_Header *msg;
void       *ctx;
{
SUT_Connection *conn = jobs->conn + i;

conn->stat     = msg->len;
jobs->retstat |= msg->len;
msg->op	       = SUT_EXIT;
WRITE( conn->writefd, msg, sizeof(SUT_Header) );

return SUT_SUCCESS;
}

/* Handle an EXIT from the parent */
int SUT_Do_Exit( jobs, i, msg, ctx )
SUT_Jobs   *jobs;
int        i;
SUT_Header *msg;
void       *ctx;
{
return SUT_SUCCESS;
}

/****************************************************************************
 Handshake controls.  Sets the connection state and handles the
 transfer of data.  Not yet used.
 ****************************************************************************/
int SUT_Do_Alive( jobs, i, msg, ctx )
SUT_Jobs   *jobs;
int        i;
SUT_Header *msg;
void       *ctx;
{
SUT_Connection *conn = jobs->conn + i;
int parms[2];

conn->state = SUT_ALIVE;
msg->op	    = SUT_PROCEED;
msg->len    = 2 * sizeof(int);
WRITE( conn->writefd, msg, sizeof(SUT_Header) );
parms[0]    = SUT_Debug;
parms[1]    = SUT_Options;
WRITE( conn->writefd, parms, 2 * sizeof(int) );

return SUT_SUCCESS;
}

int SUT_Do_Proceed( jobs, i, msg, ctx )
SUT_Jobs   *jobs;
int        i;
SUT_Header *msg;
void       *ctx;
{
SUT_Connection *conn = jobs->conn + i;
int parms[2];
READ( conn->readfd, parms, 2 * sizeof(int) );
conn->state = SUT_STARTED;
SUT_Debug	= parms[0];
SUT_Options	= parms[1];
SUT_Mynode  = SUT_Mynodenum();

return SUT_SUCCESS;
}

/****************************************************************************/
/* Special function for 'undefined' */
int SUT_OpUnknown( jobs, i, msg, ctx )
SUT_Jobs   *jobs;
int        i;
SUT_Header *msg;
void       *ctx;
{
fprintf( stderr, "Unknown op field %d\n", msg->op );
/* If the op is in range, we should try to give its name */
return SUT_SUCCESS;
}

/****************************************************************************
 Create and manage the SUT_Opfcns structure
 ****************************************************************************/
SUT_Opfcns *SUT_CreateOpfcns()
{
SUT_Opfcns *new;
int        i;

new = (SUT_Opfcns *)MALLOC( sizeof(SUT_Opfcns) );
if (!new) return 0;
for (i=0; i<SUT_MAXOP; i++) 
    new->msghandler[i] = SUT_OpUnknown;
new->msghandler[SUT_STDOUT]  = SUT_Data_from_child;
new->msghandler[SUT_DONE]    = SUT_Do_Done;
new->msghandler[SUT_EXIT]    = SUT_Do_Exit;

/* These will eventually handle the startup handshake ... */
new->msghandler[SUT_ALIVE]   = SUT_Do_Alive;
new->msghandler[SUT_PROCEED] = SUT_Do_Proceed;

return new;
}
/****************************************************************************
 These are the service routines called by the active-message handlers
 ****************************************************************************/

/* read data from an fd.  Read the specified amount; does multiple reads
   if necessary */
int SUT_Readdata( fd, buf, len )
int  fd, len;
char *buf;
{
char *p = buf;
int  rlen;

while (len) {
    rlen = read( fd, p, len );
    if (rlen == 0) 
	return SUT_UNEXPECTED_EOF;
    if (rlen < 0) 
	return SUT_READ_FAILED;
    len -= rlen;
    p   += rlen;
    }
return SUT_SUCCESS;
}
