#include "sut.h"

/*
   This file contains the routines that manage data flow between nodes of the
   tree.
*/

/* Write data to an fd.  Use the write protocol */
int SUT_Writedata( fd, buf, len )
int fd, len;
char *buf;
{
SUT_Header head;

head.op  = SUT_STDOUT;
head.len = len;
WRITE(fd, &head, sizeof(SUT_Header));
WRITE(fd, buf, len );

if (SUT_Debug & SUT_PASSDATA) {
    buf[len] = 0;
    fprintf( stderr, "[%d] Writing (%x)|%s| in Writedata\n", 
	     SUT_Mynode, buf[0], buf );
    }
return SUT_SUCCESS;
}

/* At the root, there is no one to interpret the "header" */
int SUT_Writedataroot( fd, buf, len )
int fd, len;
char *buf;
{
SUT_Header head;

WRITE(fd, buf, len );

if (SUT_Debug & SUT_PASSDATA) {
    buf[len] = 0;
    fprintf( stderr, "[%d] Writing (%x)|%s| in Writedataroot\n", 
	     SUT_Mynode, buf[0], buf );
    }
return SUT_SUCCESS;
}

/* This is a handshake routine to tell child that it can write to stdout */
int SUT_AllowWrites( jobs, i )
SUT_Jobs *jobs;
int      i;
{
SUT_Header buf;
SUT_Connection *conn = jobs->conn + i;

if (SUT_Debug & SUT_PASSDATA) 
    fprintf( stderr, "[%d] Sending write-ok to %d\n", SUT_Mynode, 
	     conn->nodenum );
buf.op = SUT_ALLOW;
WRITE( conn->writefd, &buf, sizeof(SUT_Header) );
return SUT_SUCCESS;
}

/* Check to see that data is actually available, in case a select 
   improperly set a descriptor */
int SUT_Data_available( fd )
int fd;
{
struct timeval timer;
fd_set         rd_ready;
timer.tv_sec  = 0;
timer.tv_usec = 0;
/* If select didn't lie about what was ready, we would
   not have to do this... */
FD_ZERO(&rd_ready);
FD_SET(fd, &rd_ready );
return select( fd + 1, &rd_ready, (fd_set *)0, (fd_set *)0, &timer );
}
/*

Write data (buf,len) for job i to stdout.  Buffer if there isn't enough

 */
int SUT_Writebuffered( jobs, i, fd )
SUT_Jobs *jobs;
int      i, fd;
{
SUT_Connection *conn = jobs->conn + i;

if (conn->linepos && jobs->availout >= conn->linepos) {
    if (SUT_Debug & SUT_PASSDATA) 
	fprintf( stderr, 
		 "[%d] Writing data to node %d (%d bytes) in Writebuffered\n", 
		 SUT_Mynode, conn->nodenum, conn->linepos );
    (*jobs->writeparent)( fd, conn->linebuf, conn->linepos );
    jobs->availout -= conn->linepos;
    conn->linepos  = 0;
    /* Send to this child that we can now accept input */
    SUT_AllowWrites( jobs, i );
    }
/* Do nothing if we can't output */
return SUT_SUCCESS;
}

/*
   Read from a child and write to stdout (1).  Calls specified writer
   routine.
 */
int SUT_Read_child( jobs, i )
SUT_Jobs *jobs;
int      i;
{
int            len;
SUT_Connection *conn = jobs->conn + i;

if (SUT_Debug & SUT_PASSDATA)
    fprintf( stderr, "[%d] About to read %dth job, node %d\n", 
	     SUT_Mynode, i, conn->nodenum );
if (SUT_Data_available( conn->readfd ) == 1) {
    if (SUT_Debug & SUT_PASSDATA)
	fprintf( stderr, 
		"[%d] About to read %dth job (poll succeeded)\n", 
		SUT_Mynode, i );
    len = (*conn->readdata)( jobs, i );
    if (len <= 0) {
	/* Need to detect EOF and remove from list of 
	   active nodes */
	if (SUT_Debug & SUT_PASSDATA) 
	    fprintf( stderr, "[%d] Eof on read\n", SUT_Mynode );
	SUT_Do_job_exit( i, jobs );
	}
    if (SUT_Debug & SUT_PASSDATA)
	fprintf( stderr, "[%d] End of read\n", SUT_Mynode );
    return (len == 0) ? SUT_CHILD_DONE : SUT_SUCCESS;
    }
else {
    if (SUT_Debug & SUT_PASSDATA)
	fprintf( stderr, "[%d] Select LIED!\n", SUT_Mynode );
    }
return SUT_SUCCESS;
}

/* @
   SUT_Passdata - Pass data to and from stdin/stdout/stderr
   
   See also:
   SUT_Regulate_data - Pass data, but with a throttle
@ */
int SUT_Passdata( jobs )
SUT_Jobs *jobs;
{
int            i, navail, nfds, len, njobs, err;
char           buf[SUT_MAX_BUF];
fd_set         rd_ready;
struct timeval timer;
SUT_Connection *conn;

nfds    = jobs->maxfds;
njobs   = jobs->nconn;
while (jobs->nactive) {
    /* We need to read from the pipes for each of the created processes */
    navail = select( nfds+1, &jobs->rdset, (fd_set *)0, (fd_set *)0, 
		    (struct timeval *)0 );
    /* fprintf( stderr, "navail = %d, errno = %d\n", navail, errno ); */
    /* We'd like to do (if (navail < 0 && errno == EINTR) continue,
       but on some systems (SunOS for one), there may be data waiting
       despite the description in the manual.  So we just check
       for everything anyway ... */
    for (i=0; i<njobs; i++) {
	conn = jobs->conn + i;
	if (conn->readfd >= 0) {
	    /* Look for data first */
	    if (FD_ISSET(conn->readfd,&jobs->rdset)) {
		if (err = SUT_Read_child( jobs, i )) {
		    if (err != SUT_CHILD_DONE) 
			SUT_Errormessage( stderr, err );
		    }
		}
	    else
		FD_SET(conn->readfd,&jobs->rdset);
	    }
	}
    }
return SUT_SUCCESS;
}

/* @
   SUT_Regulate_data - Pass data to and from stdin/stdout/stderr, controlling
   the volume of data
   
   See also:
   SUT_Passdata - Pass data, but without a throttle
@ */
int SUT_Regulate_data( jobs )
SUT_Jobs *jobs;
{
int            i, navail, nfds, len, njobs, err;
char           buf[SUT_MAX_BUF];
fd_set         rd_ready;
SUT_Connection *conn;
struct timeval timer;

nfds    = jobs->maxfds;
njobs   = jobs->nconn;
jobs->writefilter = SUT_Writebuffered;
/* Look at stdin for flow control */
FD_SET( 0, &jobs->rdset );
while (jobs->nactive) {
    /* We need to read from the pipes for each of the created processes */
    navail = select( nfds+1, &jobs->rdset, (fd_set *)0, (fd_set *)0,
		    (struct timeval *)0 );
    /* fprintf( stderr, "navail = %d, errno = %d\n", navail, errno ); */
    /* We'd like to do (if (navail < 0 && errno == EINTR) continue,
       but on some systems (SunOS for one), there may be data waiting
       despite the description in the manual.  So we just check
       for everything anyway ... */
    if (FD_ISSET( 0, &jobs->rdset )) {
	/* Command from stdin */
	if (SUT_Debug & SUT_PASSDATA) 
	    fprintf( stderr, "[%d] About to read from stdin\n", SUT_Mynode );
	/* For now, always flow control */
	len = 0;
	if (SUT_Data_available( 0 ) == 1) {
	    len = read( 0, buf, SUT_MAX_BUF );
	    if (len > 0) {
		if (SUT_Debug & SUT_PASSDATA) 
		    fprintf( stderr, "[%d] Read %d bytes from stdin\n", 
			     SUT_Mynode, len );
		jobs->availout = SUT_MAX_BUF;
		if (SUT_Debug & SUT_PASSDATA) {
		    if (isascii(buf[0])) {
			buf[len] = 0;
			fprintf( stderr, "[%d] Read %s from stdin\n", 
				 SUT_Mynode, buf );
			}
		    else {
			fprintf( stderr, "[%d] Read %x from stdin\n", 
				 SUT_Mynode, buf[0] );
			}
		    }
		}
	    else {
		if (len == 0) 
		    fprintf( stderr, "[%d] EOF on input\n", SUT_Mynode );
		}
	    }
	}
    else 
	FD_SET( 0, &jobs->rdset );
    for (i=0; i<njobs; i++) {
	conn = jobs->conn + i;
	if (conn->readfd >= 0) {
	    /* Look for data first */
	    if (FD_ISSET(conn->readfd,&jobs->rdset)) {
		if (err = SUT_Read_child( jobs, i )) {
		    if (err != SUT_CHILD_DONE) 
			SUT_Errormessage( stderr, err );
		    }
		}
	    else
		FD_SET(conn->readfd,&jobs->rdset);
	    }
	}
    }
return SUT_SUCCESS;
}

/*
   Once we've reached the final form for these routines, we want to add two
   more:

   OutputInOrder - read from lowest numbered until EOF, then read the
   next, and so on.  Good for generating node-number sorted output

   OutputCombined - read and merge, so that common lines are listed as
   coming from multiple nodes
 */
